# 
# 
# add_file "configs/smdi.conf.sample"
# 
# add_file "doc/README.smdi"
# 
# add_file "include/asterisk/smdi.h"
# 
# add_file "res/res_smdi.c"
# 
# patch "Makefile"
#  from [14a6a326c6c9fa8462fa8c07018e037884b3fc97]
#    to [2c70aa1d376e73d5b9893226df351396fd45738f]
# 
# patch "apps/Makefile"
#  from [41ce1c1b5958837b6618b84e13cd3171677d9283]
#    to [9be45c7114ba536100203574789afb61838a4475]
# 
# patch "apps/app_voicemail.c"
#  from [40aadc8466eccf1765830cd069ca84204041563f]
#    to [0bc9c5f1e086a4eaab4ab6229459ff0c7ab3d043]
# 
# patch "channels/Makefile"
#  from [e9e74701144180a2300458827452e5bd785a59c3]
#    to [cd63dfa3debbdbc9cf27147aa99597469db8f42c]
# 
# patch "channels/chan_zap.c"
#  from [c363d7352632cbf7286fcec3328392a72dfa7a0b]
#    to [4357ca101d035d8913ad3411e3539f9814ff392c]
# 
# patch "configs/smdi.conf.sample"
#  from []
#    to [5466a78a9e48309a3cb2f1ea7527f7f9cbbfeb9e]
# 
# patch "configs/voicemail.conf.sample"
#  from [4602c00c3940f791c133da1b68319b99f2032984]
#    to [dfe10ad643181c9fe72a6f87eeb4f44483c4f6a7]
# 
# patch "configs/zapata.conf.sample"
#  from [077f657f1c4612da8098e7b7a6139f57fd48ff72]
#    to [99e6b3f9af1ae7916eff845c7d2b73c0275d2a8f]
# 
# patch "doc/README.smdi"
#  from []
#    to [3923f6e2a3d5c25288b6390981d50cce0b448db2]
# 
# patch "include/asterisk/astobj.h"
#  from [15f1bb8aff7b84f369edb41f9b11cce52f8a630e]
#    to [0d3dc8aae2d3fffeeb4eb7a863d12ffbf622d4ce]
# 
# patch "include/asterisk/callerid.h"
#  from [30683d737dbddb00385afa5590fa2c779de9f290]
#    to [22f4cdff4a98ef2d8a61ea0b23d77e74147adc62]
# 
# patch "include/asterisk/smdi.h"
#  from []
#    to [c51a622b665d35b2972df9a61fad6590759da850]
# 
# patch "res/Makefile"
#  from [abfddf470e2698ce381858ea65fdec55bf8f0300]
#    to [795f85dd3c09aa80008705e0bfcb05c7811fa950]
# 
# patch "res/res_smdi.c"
#  from []
#    to [aca03f145b0b867ee9aa7515620d76ed9b664261]
# 
============================================================
--- Makefile	14a6a326c6c9fa8462fa8c07018e037884b3fc97
+++ Makefile	2c70aa1d376e73d5b9893226df351396fd45738f
@@ -68,6 +68,11 @@
 #OPTIONS += -DLOW_MEMORY
 #endif
 
+#
+# Asterisk SMDI integration
+#
+#WITH_SMDI = 1
+
 # Optional debugging parameters
 DEBUG_THREADS = #-DDUMP_SCHEDULER #-DDEBUG_SCHEDULER #-DDEBUG_THREADS #-DDO_CRASH #-DDETECT_DEADLOCKS
 
============================================================
--- apps/Makefile	41ce1c1b5958837b6618b84e13cd3171677d9283
+++ apps/Makefile	9be45c7114ba536100203574789afb61838a4475
@@ -32,6 +32,13 @@
      app_mixmonitor.so app_stack.so
 
 #
+# Asterisk SMDI integration
+#
+ifeq (${WITH_SMDI},1)
+CFLAGS+=-DWITH_SMDI
+endif
+
+#
 # Obsolete things...
 #
 #APPS+=app_sql_postgres.so
============================================================
--- apps/app_voicemail.c	40aadc8466eccf1765830cd069ca84204041563f
+++ apps/app_voicemail.c	0bc9c5f1e086a4eaab4ab6229459ff0c7ab3d043
@@ -70,6 +70,9 @@
 #include "asterisk/localtime.h"
 #include "asterisk/cli.h"
 #include "asterisk/utils.h"
+#ifdef WITH_SMDI
+#include "asterisk/smdi.h"
+#endif
 #ifdef USE_ODBC_STORAGE
 #include "asterisk/res_odbc.h"
 #endif
@@ -382,7 +385,9 @@
 static char serveremail[80];
 static char mailcmd[160];	/* Configurable mail cmd */
 static char externnotify[160]; 
-
+#ifdef WITH_SMDI
+static struct ast_smdi_interface *smdi_iface = NULL;
+#endif
 static char vmfmts[80];
 static int vmminmessage;
 static int vmmaxmessage;
@@ -2316,13 +2321,34 @@
 	char arguments[255];
 	char ext_context[256] = "";
 	int newvoicemails = 0, oldvoicemails = 0;
+#ifdef WITH_SMDI
+	struct ast_smdi_mwi_message *mwi_msg;
+#endif
 
 	if (!ast_strlen_zero(context))
 		snprintf(ext_context, sizeof(ext_context), "%s@%s", extension, context);
 	else
 		ast_copy_string(ext_context, extension, sizeof(ext_context));
 
+#ifdef WITH_SMDI
+	if (!strcasecmp(externnotify, "smdi")) {
+
+		if(ast_app_has_voicemail(ext_context, NULL)) 
+			ast_smdi_mwi_set(smdi_iface, extension);
+		else
+			ast_smdi_mwi_unset(smdi_iface, extension);
+
+		mwi_msg = ast_smdi_mwi_message_wait(smdi_iface, 1000);
+		if(mwi_msg) {
+			ast_log(LOG_ERROR, "Error executing SMDI MWI change for %s on %s\n", extension, smdi_iface->name);
+			ASTOBJ_UNREF(mwi_msg, ast_smdi_mwi_message_destroy);
+		} else {
+			ast_log(LOG_DEBUG, "Successfully executed SMDI MWI change for %s on %s\n", extension, smdi_iface->name);
+		}
+	} else if (!ast_strlen_zero(externnotify)) {
+#else
 	if (!ast_strlen_zero(externnotify)) {
+#endif
 		if (messagecount(ext_context, &newvoicemails, &oldvoicemails)) {
 			ast_log(LOG_ERROR, "Problem in calculating number of voicemail messages available for extension %s\n", extension);
 		} else {
@@ -5811,6 +5837,9 @@
 	char *cat;
 	struct ast_variable *var;
 	char *notifystr = NULL;
+#ifdef WITH_SMDI
+	char *smdistr = NULL;
+#endif
 	char *astattach;
 	char *astsearch;
 	char *astsaycid;
@@ -5920,6 +5949,25 @@
 		if ((notifystr = ast_variable_retrieve(cfg, "general", "externnotify"))) {
 			ast_copy_string(externnotify, notifystr, sizeof(externnotify));
 			ast_log(LOG_DEBUG, "found externnotify: %s\n", externnotify);
+#ifdef WITH_SMDI
+			if(!strcasecmp(externnotify, "smdi")) {
+				ast_log(LOG_DEBUG, "Using SMDI for external voicemail notification\n");
+
+				if ((smdistr = ast_variable_retrieve(cfg, "general", "smdiport"))) {
+					smdi_iface = ast_smdi_interface_find(smdistr);
+				} else {
+					ast_log(LOG_DEBUG, "No SMDI interface set, trying default (/dev/ttyS0)\n");
+					smdi_iface = ast_smdi_interface_find("/dev/ttyS0");
+				}
+
+				if(!smdi_iface) {
+					ast_log(LOG_ERROR, "No valid SMDI interface specfied, disabling external voicemail notification\n");
+					externnotify[0] = '\0';
+				} else {
+					ast_log(LOG_DEBUG, "Using SMDI port %s\n", smdi_iface->name);
+				}
+			}
+#endif
 		} else {
 			externnotify[0] = '\0';
 		}
============================================================
--- channels/Makefile	e9e74701144180a2300458827452e5bd785a59c3
+++ channels/Makefile	cd63dfa3debbdbc9cf27147aa99597469db8f42c
@@ -77,6 +77,13 @@
   CHANNEL_LIBS+=chan_phone.so
 endif
 
+#
+# Asterisk SMDI integration
+#
+ifeq (${WITH_SMDI},1)
+CFLAGS+=-DWITH_SMDI
+endif
+
 ifneq ($(wildcard h323/libchanh323.a),)
   CHANNEL_LIBS+=chan_h323.so
 endif
============================================================
--- channels/chan_zap.c	c363d7352632cbf7286fcec3328392a72dfa7a0b
+++ channels/chan_zap.c	4357ca101d035d8913ad3411e3539f9814ff392c
@@ -96,6 +96,10 @@
 #include "asterisk/term.h"
 #include "asterisk/utils.h"
 #include "asterisk/transcap.h"
+#ifdef WITH_SMDI
+#include "asterisk/smdi.h"
+#include "asterisk/astobj.h"
+#endif
 
 #ifndef ZT_SIG_EM_E1
 #error "Your zaptel is too old.  please cvs update"
@@ -270,7 +274,10 @@
 static int amaflags = 0;
 
 static int adsi = 0;
-
+#ifdef WITH_SMDI
+static int use_smdi = 0;
+static char smdi_port[SMDI_MAX_FILENAME_LEN] = "/dev/ttyS0";
+#endif
 static int numbufs = 4;
 
 static int cur_prewink = -1;
@@ -596,6 +603,10 @@
 	unsigned int r2blocked:1;
 	unsigned int sigchecked:1;
 #endif
+#ifdef WITH_SMDI
+	unsigned int use_smdi:1;		/* Whether to use SMDI on this channel */
+	struct ast_smdi_interface *smdi_iface;	/* The serial port to listen for SMDI data on */
+#endif
 
 	struct zt_distRings drings;
 
@@ -2149,6 +2160,10 @@
 		p->prev->next = p->next;
 	if(p->next)
 		p->next->prev = p->prev;
+#ifdef WITH_SMDI
+	if(p->use_smdi)
+		ASTOBJ_UNREF(p->smdi_iface, ast_smdi_interface_destroy);
+#endif
 	ast_mutex_destroy(&p->lock);
 	free(p);
 	*pvt = NULL;
@@ -5219,7 +5234,7 @@
 	unsigned char buf[256];
 	char dtmfcid[300];
 	char dtmfbuf[300];
-	struct callerid_state *cs;
+	struct callerid_state *cs = NULL;
 	char *name=NULL, *number=NULL;
 	int distMatches;
 	int curRingData[3];
@@ -5227,7 +5242,9 @@
 	int counter1;
 	int counter;
 	int samples = 0;
-
+#ifdef WITH_SMDI
+	struct ast_smdi_md_message *smdi_msg = NULL;
+#endif
 	int flags;
 	int i;
 	int timeout;
@@ -5826,10 +5843,35 @@
 			}
 		}
 #endif
+#ifdef WITH_SMDI
+		/* check for SMDI messages */
+		if(p->use_smdi && p->smdi_iface) {
+			smdi_msg = ast_smdi_md_message_wait(p->smdi_iface, 1500);
+
+			if(smdi_msg != NULL) {
+				ast_copy_string(chan->exten, smdi_msg->fwd_st, sizeof(chan->exten));
+
+				if (smdi_msg->type == 'B')
+					pbx_builtin_setvar_helper(chan, "_SMDI_VM_TYPE", "b");
+				else if (smdi_msg->type == 'N')
+					pbx_builtin_setvar_helper(chan, "_SMDI_VM_TYPE", "u");
+
+				ast_log(LOG_DEBUG, "Recieved SMDI message on %s\n", chan->name);
+			} else {
+				ast_log(LOG_WARNING, "SMDI enabled but no SMDI message present\n");
+			}
+		}
+
+		if (p->use_callerid && (p->cid_signalling == CID_SIG_SMDI && smdi_msg)) {
+	     		number = smdi_msg->calling_st;
+
 		/* If we want caller id, we're in a prering state due to a polarity reversal
 		 * and we're set to use a polarity reversal to trigger the start of caller id,
 		 * grab the caller id and wait for ringing to start... */
+		} else if (p->use_callerid && (chan->_state == AST_STATE_PRERING && p->cid_start == CID_START_POLARITY)) {
+#else
 		if (p->use_callerid && (chan->_state == AST_STATE_PRERING && p->cid_start == CID_START_POLARITY)) {
+#endif
 			/* If set to use DTMF CID signalling, listen for DTMF */
 			if (p->cid_signalling == CID_SIG_DTMF) {
 				int i = 0;
@@ -6171,7 +6213,10 @@
 			ast_shrink_phone_number(number);
 
 		ast_set_callerid(chan, number, name, number);
-
+#ifdef WITH_SMDI
+		if (smdi_msg)
+			ASTOBJ_UNREF(smdi_msg, ast_smdi_md_message_destroy);
+#endif
 		if (cs)
 			callerid_free(cs);
 		ast_setstate(chan, AST_STATE_RING);
@@ -7148,6 +7193,9 @@
 		tmp->callwaitingcallerid = callwaitingcallerid;
 		tmp->threewaycalling = threewaycalling;
 		tmp->adsi = adsi;
+#ifdef WITH_SMDI
+		tmp->use_smdi = use_smdi;
+#endif
 		tmp->permhidecallerid = hidecallerid;
 		tmp->callreturn = callreturn;
 		tmp->echocancel = echocancel;
@@ -7180,6 +7228,22 @@
 			}
 		}
 
+#ifdef WITH_SMDI
+		if (tmp->cid_signalling == CID_SIG_SMDI) {
+			if (!tmp->use_smdi) {
+				ast_log(LOG_WARNING, "SMDI callerid requires SMDI to be enabled, enabling...\n");
+				tmp->use_smdi = 1;
+			}
+		}
+		if (tmp->use_smdi) {
+			tmp->smdi_iface = ast_smdi_interface_find(smdi_port);
+			if (!(tmp->smdi_iface)) {
+				ast_log(LOG_ERROR, "Invalid SMDI port specfied, disabling SMDI support\n");
+				tmp->use_smdi = 0;
+			}
+		}
+#endif
+
 		ast_copy_string(tmp->accountcode, accountcode, sizeof(tmp->accountcode));
 		tmp->amaflags = amaflags;
 		if (!here) {
@@ -10332,6 +10396,10 @@
 				cid_signalling = CID_SIG_V23;
 			else if (!strcasecmp(v->value, "dtmf"))
 				cid_signalling = CID_SIG_DTMF;
+#ifdef WITH_SMDI
+			else if (!strcasecmp(v->value, "smdi"))
+				cid_signalling = CID_SIG_SMDI;
+#endif
 			else if (ast_true(v->value))
 				cid_signalling = CID_SIG_BELL;
 		} else if (!strcasecmp(v->name, "cidstart")) {
@@ -10354,6 +10422,12 @@
 			ast_copy_string(mailbox, v->value, sizeof(mailbox));
 		} else if (!strcasecmp(v->name, "adsi")) {
 			adsi = ast_true(v->value);
+#ifdef WITH_SMDI
+		} else if (!strcasecmp(v->name, "usesmdi")) {
+			use_smdi = ast_true(v->value);
+		} else if (!strcasecmp(v->name, "smdiport")) {
+			ast_copy_string(smdi_port, v->value, sizeof(smdi_port));
+#endif
 		} else if (!strcasecmp(v->name, "transfer")) {
 			transfer = ast_true(v->value);
 		} else if (!strcasecmp(v->name, "canpark")) {
============================================================
--- configs/smdi.conf.sample	
+++ configs/smdi.conf.sample	5466a78a9e48309a3cb2f1ea7527f7f9cbbfeb9e
@@ -0,0 +1,46 @@
+; Asterisk SMDI configuration
+
+[interfaces]
+; Specify serial ports to listen for SMDI messages on below.  These will be
+; referenced later in zapata.conf.  If you do not specify any interfaces then
+; SMDI will be disabled.  Interfaces can have several different attributes
+; associated with them.
+
+; Configure the type of switch you will be connecting with.  This setting will
+; effect how Asterisk handles SMDI messages.
+; 
+; Valid options are:
+;    default  = SMDI according to Bell spec TSR-TSY-000283
+;    dms10    = Nortel DMS10
+;    g3       = AT&T/Lucent G3
+
+; switchtype = dms10
+
+; Set the number of stop bits to use per character here.  The default is no,
+; in which case one stop bit will be used.
+
+;twostopbits = no
+
+; Character size or bit length is the size of each character sent accross the
+; link.  Character size can be 7 or 8.  The default is 7.
+
+;charsize = 7
+
+; If you need parity checking enabled you can turn it on here.  Acceptable
+; values are even, odd, and none.  The default is even.
+
+;paritybit = even
+
+; The baudrate to use for this port.  Acceptable values are 1200, 2400, 4800,
+; and 9600.  The default is 9600.
+
+;baudrate = 1200
+
+; Often the numbering scheme for a set of mailboxes or extensions will not be 7
+; or 10 digits (as SMDI requires).  Use the msdstrip option to strip unused
+; digits from the start of numbers.
+
+;msdstrip = 0
+
+;smdiport => /dev/ttyS0
+
============================================================
--- configs/voicemail.conf.sample	4602c00c3940f791c133da1b68319b99f2032984
+++ configs/voicemail.conf.sample	dfe10ad643181c9fe72a6f87eeb4f44483c4f6a7
@@ -49,12 +49,18 @@
 silencethreshold=128
 ; Max number of failed login attempts
 maxlogins=3
-; If you need to have an external program, i.e. /usr/bin/myapp called when a
-; voicemail is left, delivered, or your voicemailbox is checked, uncomment
-; this:
+
+; If you need to have an external program, i.e. /usr/bin/myapp
+; called when a voicemail is left, delivered, or your voicemailbox 
+; is checked, uncomment this.  It can also be set to 'smdi' to use
+; smdi for external notification.  If it is 'smdi', smdiport should
+; be set to a valid port as specfied in smdi.conf.
+
 ;externnotify=/usr/bin/myapp
-; If you need to have an external program, i.e. /usr/bin/myapp called when a
-; voicemail password is changed, uncomment this:
+;smdiport=/dev/ttyS0
+
+; If you need to have an external program, i.e. /usr/bin/myapp
+; called when a voicemail password is changed, uncomment this:
 ;externpass=/usr/bin/myapp
 ; For the directory, you can override the intro file if you want
 ;directoryintro=dir-intro
============================================================
--- configs/zapata.conf.sample	077f657f1c4612da8098e7b7a6139f57fd48ff72
+++ configs/zapata.conf.sample	99e6b3f9af1ae7916eff845c7d2b73c0275d2a8f
@@ -222,6 +222,7 @@
 ;     bell     = bell202 as used in US
 ;     v23      = v23 as used in the UK
 ;     dtmf     = DTMF as used in Denmark, Sweden and Netherlands
+;     smdi     = Use SMDI for callerid.  Requires SMDI to be enabled (usesmdi).
 ;
 ;cidsignalling=bell
 ;
@@ -376,6 +377,14 @@
 ;
 ;adsi=yes
 ;
+; SMDI (Simplified Message Desk Interface) can be enabled on a per-channel
+; basis if you would like that channel to behave like an SMDI message desk.
+; The SMDI port specfied should have already been defined in smdi.conf.  The
+; default port is /dev/ttyS0.
+;
+;usesmdi=yes
+;smdiport=/dev/ttyS0
+;
 ; On trunk interfaces (FXS) and E&M interfaces (E&M, Wink, Feature Group D
 ; etc, it can be useful to perform busy detection either in an effort to 
 ; detect hangup or for detecting busies.  This enables listening for
============================================================
--- doc/README.smdi	
+++ doc/README.smdi	3923f6e2a3d5c25288b6390981d50cce0b448db2
@@ -0,0 +1,34 @@
+Asterisk SMDI (Simple Message Desk Interface) integration
+---------------------------------------------------------
+
+Support for using Asterisk as an SMDI message desk was developed by Matthew A.
+Nicholson <mnicholson@digium.com>.  To enable SMDI support in asterisk edit the
+Makefile file and uncomment the line regarding SMDI so it reads like
+this:
+
+#
+# Asterisk SMDI integration
+#
+WITH_SMDI = 1
+
+SMDI integration is configured in smdi.conf, zaptel.conf, and voicemail.conf.
+Various characteristics of the SMDI interfaces to be used (serial ports) are
+defined in smdi.conf.  SMDI integration for callerid and MWI are defined in
+zaptel.conf and voicemail.conf respectively.
+
+When SMDI is enabled and a call comes into Asterisk, the forwarding station
+number is used as the destination for the call and any callerid information
+present is used.  This way you can configure your extensions.conf as follows to
+behave as a message desk.
+
+[default]
+
+exten => _XXXXXXX,1,VoiceMail(${EXTEN}|${SMDI_VM_TYPE})
+exten => _XXXXXXX,n,Hangup
+
+exten => s,1,VoiceMailMain(${CALLERIDNUM})
+exten => s,n,Hangup
+
+The ${SMDI_VM_TYPE} variable will be set to u, b, or nothing depending on the
+contents of the type of SMDI message received.
+
============================================================
--- include/asterisk/astobj.h	15f1bb8aff7b84f369edb41f9b11cce52f8a630e
+++ include/asterisk/astobj.h	0d3dc8aae2d3fffeeb4eb7a863d12ffbf622d4ce
@@ -571,6 +571,76 @@
 		found; \
 	})
 
+/*! \brief Add an object to the end of a container.
+ *
+ * \param container A pointer to the container to operate on.
+ * \param newobj A pointer to the object to be added.
+ *
+ * This macro adds an object to the end of a container.
+ */
+#define ASTOBJ_CONTAINER_LINK_END(container,newobj) \
+	do { \
+		typeof((container)->head) iterator; \
+		typeof((container)->head) next; \
+		typeof((container)->head) prev; \
+		ASTOBJ_CONTAINER_RDLOCK(container); \
+		prev = NULL; \
+		next = (container)->head; \
+		while((iterator = next)) { \
+			next = iterator->next[0]; \
+			prev = iterator; \
+		} \
+		if(prev) { \
+			ASTOBJ_CONTAINER_WRLOCK((container)); \
+			prev->next[0] = ASTOBJ_REF(newobj); \
+			(newobj)->next[0] = NULL; \
+			ASTOBJ_CONTAINER_UNLOCK((container)); \
+		} else { \
+			ASTOBJ_CONTAINER_LINK_START((container),(newobj)); \
+		} \
+		ASTOBJ_CONTAINER_UNLOCK((container)); \
+	} while(0)
+
+/*! \brief Add an object to the front of a container.
+ *
+ * \param container A pointer to the container to operate on.
+ * \param newobj A pointer to the object to be added.
+ *
+ * This macro adds an object to the start of a container.
+ */
+#define ASTOBJ_CONTAINER_LINK_START(container,newobj) \
+	do { \
+		ASTOBJ_CONTAINER_WRLOCK(container); \
+		(newobj)->next[0] = (container)->head; \
+		(container)->head = ASTOBJ_REF(newobj); \
+		ASTOBJ_CONTAINER_UNLOCK(container); \
+	} while(0)
+
+/*! \brief Remove an object from the front of a container.
+ *
+ * \param container A pointer to the container to operate on.
+ *
+ * This macro removes the first object in a container.
+ *
+ * \note This macro does not destroy any objects, it simply unlinks
+ * them from the list.  No destructors are called.
+ *
+ * \return The container's reference to the removed object or NULL if no
+ * matching object was found.
+ */
+#define ASTOBJ_CONTAINER_UNLINK_START(container) \
+	({ \
+		typeof((container)->head) found = NULL; \
+		ASTOBJ_CONTAINER_WRLOCK(container); \
+	 	if((container)->head) { \
+	 		found = (container)->head; \
+	 		(container)->head = (container)->head->next[0]; \
+	 		found->next[0] = NULL; \
+	 	} \
+		ASTOBJ_CONTAINER_UNLOCK(container); \
+	 	found; \
+	})
+
 /*! \brief Prune marked objects from a container.
  *
  * \param container A pointer to the container to prune.
============================================================
--- include/asterisk/callerid.h	30683d737dbddb00385afa5590fa2c779de9f290
+++ include/asterisk/callerid.h	22f4cdff4a98ef2d8a61ea0b23d77e74147adc62
@@ -41,6 +41,9 @@
 #define CID_SIG_BELL	1
 #define CID_SIG_V23	2
 #define CID_SIG_DTMF	3
+#ifdef WITH_SMDI
+#define CID_SIG_SMDI	4
+#endif
 
 #define CID_START_RING	1
 #define CID_START_POLARITY 2
============================================================
--- include/asterisk/smdi.h	
+++ include/asterisk/smdi.h	c51a622b665d35b2972df9a61fad6590759da850
@@ -0,0 +1,118 @@
+/*
+ * Asterisk -- A telephony toolkit for Linux.
+ *
+ * SMDI support for Asterisk.
+ * 
+ * Copyright (C) 2005, Digium, Inc.
+ *
+ * Matthew A. Nicholson <mnicholson@digium.com>
+ *
+ * This program is free software, distributed under the terms of
+ * the GNU General Public License.
+ */
+
+/*! 
+ * \file
+ * \brief SMDI support for Asterisk.
+ */
+
+
+/* C is simply a ego booster for those who want to do objects the hard way. */
+
+
+#ifndef AST_SMDI_H
+#define AST_SMDI_H
+
+#include "asterisk/config.h"
+#include "asterisk/module.h"
+#include "asterisk/astobj.h"
+#include <termios.h>
+
+#define SMDI_MESG_DESK_NUM_LEN 3
+#define SMDI_MESG_DESK_TERM_LEN 4
+#define SMDI_MWI_FAIL_CAUSE_LEN 3
+#define SMDI_MAX_STATION_NUM_LEN 10
+#define SMDI_MAX_FILENAME_LEN 256
+
+struct ast_smdi_switchtype;
+
+/*!
+ * \brief An SMDI message waiting indicator message.
+ *
+ * The ast_smdi_mwi_message structure contains the parsed out parts of an smdi
+ * message.  Each ast_smdi_interface structure has a message queue consisting
+ * ast_smdi_mwi_message structures. 
+ */
+struct ast_smdi_mwi_message {
+	ASTOBJ_COMPONENTS(struct ast_smdi_mwi_message);
+	char fwd_st[SMDI_MAX_STATION_NUM_LEN + 1];		/* forwarding station number */
+	char cause[SMDI_MWI_FAIL_CAUSE_LEN + 1];		/* the type of failure */
+};
+
+/*!
+ * \brief An SMDI message desk message.
+ *
+ * The ast_smdi_md_message structure contains the parsed out parts of an smdi
+ * message.  Each ast_smdi_interface structure has a message queue consisting
+ * ast_smdi_md_message structures. 
+ */
+struct ast_smdi_md_message {
+	ASTOBJ_COMPONENTS(struct ast_smdi_md_message);
+	char mesg_desk_num[SMDI_MESG_DESK_NUM_LEN + 1];		/* message desk number */
+	char mesg_desk_term[SMDI_MESG_DESK_TERM_LEN + 1];	/* message desk terminal */
+	char fwd_st[SMDI_MAX_STATION_NUM_LEN + 1];		/* forwarding station number */
+	char calling_st[SMDI_MAX_STATION_NUM_LEN + 1];		/* calling station number */
+	char type;						/* the type of the call */
+};
+
+/*! \brief SMDI message desk message queue. */
+struct ast_smdi_md_queue {
+	ASTOBJ_CONTAINER_COMPONENTS(struct ast_smdi_md_message);
+};
+
+/*! \brief SMDI message waiting indicator message queue. */
+struct ast_smdi_mwi_queue {
+	ASTOBJ_CONTAINER_COMPONENTS(struct ast_smdi_mwi_message);
+};
+
+/*! 
+ * \brief SMDI interface structure.
+ *
+ * The ast_smdi_interface structure holds information on a serial port that
+ * should be monitored for SMDI activity.  The structure contains a message
+ * queue of messages that have been recieved on the interface.
+ */
+struct ast_smdi_interface {
+	ASTOBJ_COMPONENTS_FULL(struct ast_smdi_interface, SMDI_MAX_FILENAME_LEN, 1);
+	struct ast_smdi_md_queue md_q;
+	struct ast_smdi_mwi_queue mwi_q;
+	FILE *file;
+	int fd;
+	pthread_t thread;
+	struct termios mode;
+	int msdstrip;
+	struct ast_smdi_switchtype *switchtype;
+};
+
+
+/* MD message queue functions */
+extern struct ast_smdi_md_message *ast_smdi_md_message_pop(struct ast_smdi_interface *iface);
+extern struct ast_smdi_md_message *ast_smdi_md_message_wait(struct ast_smdi_interface *iface, int timeout);
+extern void ast_smdi_md_message_putback(struct ast_smdi_interface *iface, struct ast_smdi_md_message *msg);
+
+/* MWI message queue functions */
+extern struct ast_smdi_mwi_message *ast_smdi_mwi_message_pop(struct ast_smdi_interface *iface);
+extern struct ast_smdi_mwi_message *ast_smdi_mwi_message_wait(struct ast_smdi_interface *iface, int timeout);
+extern void ast_smdi_mwi_message_putback(struct ast_smdi_interface *iface, struct ast_smdi_mwi_message *msg);
+
+extern struct ast_smdi_interface *ast_smdi_interface_find(const char *iface_name);
+
+/* MWI functions */
+extern int ast_smdi_mwi_set(struct ast_smdi_interface *iface, const char *mailbox);
+extern int ast_smdi_mwi_unset(struct ast_smdi_interface *iface, const char *mailbox);
+
+extern void ast_smdi_md_message_destroy(struct ast_smdi_md_message *msg);
+extern void ast_smdi_mwi_message_destroy(struct ast_smdi_mwi_message *msg);
+
+extern void ast_smdi_interface_destroy(struct ast_smdi_interface *iface);
+#endif
============================================================
--- res/Makefile	abfddf470e2698ce381858ea65fdec55bf8f0300
+++ res/Makefile	795f85dd3c09aa80008705e0bfcb05c7811fa950
@@ -69,6 +69,13 @@
 CFLAGS+=-DOPENSSL_NO_KRB5 -fPIC
 endif
 
+#
+# Asterisk SMDI integration
+#
+ifeq (${WITH_SMDI},1)
+MODS+=res_smdi.so
+endif
+
 all: depend $(MODS)
 
 install: all
============================================================
--- res/res_smdi.c	
+++ res/res_smdi.c	aca03f145b0b867ee9aa7515620d76ed9b664261
@@ -0,0 +1,865 @@
+/*
+ * Asterisk -- A telephony toolkit for Linux.
+ *
+ * SMDI support for Asterisk.
+ * 
+ * Copyright (C) 2005, Digium, Inc.
+ *
+ * Matthew A. Nicholson <mnicholson@digium.com>
+ *
+ * This program is free software, distributed under the terms of
+ * the GNU General Public License
+ */
+
+#include <stdio.h>
+#include <stdlib.h>
+#include <errno.h>
+#include <termios.h>
+#include <sys/time.h>
+#include <time.h>
+#include "asterisk/module.h"
+#include "asterisk/lock.h"
+#include "asterisk/utils.h"
+#include "asterisk/smdi.h"
+#include "asterisk/config.h"
+#include "asterisk/astobj.h"
+#include "asterisk/io.h"
+#include "asterisk/logger.h"
+#include "asterisk/utils.h"
+#include "asterisk/options.h"
+
+/*!
+ * \file
+ * \brief SMDI support for Asterisk.
+ */
+
+static const char tdesc[] = "Asterisk Simplified Message Desk Interface (SMDI) Module";
+static const char config_file[] = "smdi.conf";
+
+static void ast_smdi_md_message_push(struct ast_smdi_interface *iface, struct ast_smdi_md_message *msg);
+static void ast_smdi_mwi_message_push(struct ast_smdi_interface *iface, struct ast_smdi_mwi_message *msg);
+static void ast_smdi_switchtype_destroy(struct ast_smdi_switchtype *switchtype);
+static int ast_smdi_switchtype_add(const char *name, const char *msg_prefix, const char *md_postfix, const char *mwi_postfix, char md_cid_end);
+
+static void *smdi_read(void *iface_p);
+static int smdi_load(int reload);
+
+/* Use count stuff */
+
+AST_MUTEX_DEFINE_STATIC(localuser_lock);
+static int localusecnt = 0;
+
+/*! \brief SMDI interface container. */
+struct ast_smdi_interface_container {
+	ASTOBJ_CONTAINER_COMPONENTS(struct ast_smdi_interface);
+} smdi_ifaces;
+
+/*! 
+ * \brief SMDI switchtype container.
+ *
+ * This structure contains information about specific SMDI switchtypes so that
+ * Asterisk can adapt accordingly.  We are the Borg.
+ */
+struct ast_smdi_switchtype {
+	ASTOBJ_COMPONENTS(struct ast_smdi_switchtype);
+	char msg_prefix[8];	/* the string prefixing each SMDI message */
+	char md_postfix[8];	/* the string postfixing each MD message*/
+	char mwi_postfix[8];	/* the string postfixing each MWI message */
+	char md_cid_end;	/* the character terminating the CID string */
+};
+
+/*! \brief SMDI switchtype container. */
+static struct ast_smdi_switchtype_container {
+	ASTOBJ_CONTAINER_COMPONENTS(struct ast_smdi_switchtype);
+} smdi_switchtypes; 
+
+/*! 
+ * \brief Push an SMDI message to the back of an interface's message queue.
+ * \param iface a pointer to the interface to use.
+ * \param md_msg a pointer to the message to use.
+ */
+static void ast_smdi_md_message_push(struct ast_smdi_interface *iface, struct ast_smdi_md_message *md_msg)
+{
+	ASTOBJ_CONTAINER_LINK_END(&iface->md_q, md_msg);
+}
+
+/*!
+ * \brief Push an SMDI message to the back of an interface's message queue.
+ * \param iface a pointer to the interface to use.
+ * \param mwi_msg a pointer to the message to use.
+ */
+static void ast_smdi_mwi_message_push(struct ast_smdi_interface *iface, struct ast_smdi_mwi_message *mwi_msg)
+{
+	ASTOBJ_CONTAINER_LINK_END(&iface->mwi_q, mwi_msg);
+}
+
+/*!
+ * \brief Set the MWI indicator for a mailbox.
+ * \param iface the interface to use.
+ * \param mailbox the mailbox to use.
+ */
+extern int ast_smdi_mwi_set(struct ast_smdi_interface *iface, const char *mailbox)
+{
+	int i;
+
+	ASTOBJ_WRLOCK(iface);
+
+	fprintf(iface->file, "OP:MWI ");
+
+	for(i = 0; i < iface->msdstrip; i++)
+	   fprintf(iface->file, "0");
+
+	fprintf(iface->file, "%s!\x04", mailbox);
+
+	ASTOBJ_UNLOCK(iface);
+	ast_log(LOG_DEBUG, "Sent MWI set message for %s on %s\n", mailbox, iface->name);
+	return 0;
+}
+
+/*! 
+ * \brief Unset the MWI indicator for a mailbox.
+ * \param iface the interface to use.
+ * \param mailbox the mailbox to use.
+ */
+extern int ast_smdi_mwi_unset(struct ast_smdi_interface *iface, const char *mailbox)
+{
+	int i;
+	
+	ASTOBJ_WRLOCK(iface);
+
+	fprintf(iface->file, "RMV:MWI ");
+
+	for(i = 0; i < iface->msdstrip; i++)
+	   fprintf(iface->file, "0");
+
+	fprintf(iface->file, "%s!\x04", mailbox);
+
+	ASTOBJ_UNLOCK(iface);
+	ast_log(LOG_DEBUG, "Sent MWI unset message for %s on %s", mailbox, iface->name);
+	return 0;
+}
+
+/*!
+ * \brief Put an SMDI message back in the front of the queue.
+ * \param iface a pointer to the interface to use.
+ * \param msg a pointer to the message to use.
+ *
+ * This function puts a message back in the front of the specified queue.  It
+ * should be used if a message was popped but is not going to be processed for
+ * some reason, and the message needs to be returned to the queue.
+ */
+extern void ast_smdi_md_message_putback(struct ast_smdi_interface *iface, struct ast_smdi_md_message *md_msg)
+{
+	ASTOBJ_CONTAINER_LINK_START(&iface->md_q, md_msg);
+}
+
+/*!
+ * \brief Put an SMDI message back in the front of the queue.
+ * \param iface a pointer to the interface to use.
+ * \param msg a pointer to the message to use.
+ *
+ * This function puts a message back in the front of the specified queue.  It
+ * should be used if a message was popped but is not going to be processed for
+ * some reason, and the message needs to be returned to the queue.
+ */
+extern void ast_smdi_mwi_message_putback(struct ast_smdi_interface *iface, struct ast_smdi_mwi_message *mwi_msg)
+{
+	ASTOBJ_CONTAINER_LINK_START(&iface->mwi_q, mwi_msg);
+}
+
+/*! 
+ * \brief Get the next SMDI message from the queue.
+ * \param iface a pointer to the interface to use.
+ *
+ * This function pulls a message from the SMDI message queue on the specified
+ * interface.
+ *
+ * \return the next SMDI message, or NULL if there were no pending messages.
+ */
+extern struct ast_smdi_md_message *ast_smdi_md_message_pop(struct ast_smdi_interface *iface)
+{
+	return (ASTOBJ_CONTAINER_UNLINK_START(&iface->md_q));
+}
+
+/*!
+ * \brief Get the next SMDI message from the queue.
+ * \param iface a pointer to the interface to use.
+ * \param timeout the time to wait before returning in milliseconds.
+ *
+ * This function pulls a message from the SMDI message queue on the specified
+ * interface.  If no message is available this function will wait the specified
+ * amount of time before returning.
+ *
+ * \return the next SMDI message, or NULL if there were no pending messages and
+ * the timeout has expired.
+ */
+extern struct ast_smdi_md_message *ast_smdi_md_message_wait(struct ast_smdi_interface *iface, int timeout)
+{
+	struct timeval start, end;
+	long diff = 0;
+	struct ast_smdi_md_message *msg;
+
+	gettimeofday(&start, NULL);
+	while(diff < timeout) {
+
+		if((msg = ast_smdi_md_message_pop(iface)))
+			return msg;
+
+		/* check timeout */
+		gettimeofday(&end, NULL);
+		diff = (end.tv_sec - start.tv_sec) * 1000;
+		diff += (end.tv_usec - start.tv_usec) / 1000;
+	}
+	return (ast_smdi_md_message_pop(iface));
+}
+
+/*!
+ * \brief Get the next SMDI message from the queue.
+ * \param iface a pointer to the interface to use.
+ *
+ * This function pulls a message from the SMDI message queue on the specified
+ * interface.
+ *
+ * \return the next SMDI message, or NULL if there were no pending messages.
+ */
+extern struct ast_smdi_mwi_message *ast_smdi_mwi_message_pop(struct ast_smdi_interface *iface)
+{
+	return (ASTOBJ_CONTAINER_UNLINK_START(&iface->mwi_q));
+}
+
+/*!
+ * \brief Get the next SMDI message from the queue.
+ * \param iface a pointer to the interface to use.
+ * \param timeout the time to wait before returning in milliseconds.
+ *
+ * This function pulls a message from the SMDI message queue on the specified
+ * interface.  If no message is available this function will wait the specified
+ * amount of time before returning.
+ *
+ * \return the next SMDI message, or NULL if there were no pending messages and
+ * the timeout has expired.
+ */
+extern struct ast_smdi_mwi_message *ast_smdi_mwi_message_wait(struct ast_smdi_interface *iface, int timeout)
+{
+	struct timeval start, end;
+	long diff = 0;
+	struct ast_smdi_mwi_message *msg;
+
+	gettimeofday(&start, NULL);
+	while(diff < timeout) {
+
+		if((msg = ast_smdi_mwi_message_pop(iface)))
+			return msg;
+
+		/* check timeout */
+		gettimeofday(&end, NULL);
+		diff = (end.tv_sec - start.tv_sec) * 1000;
+		diff += (end.tv_usec - start.tv_usec) / 1000;
+	}
+	return (ast_smdi_mwi_message_pop(iface));
+}
+
+/*!
+ * \brief Find an SMDI interface with the specified name.
+ * \param iface_name the name/port of the interface to search for.
+ *
+ * \return a pointer to the interface located or NULL if none was found.  This
+ * actually returns an ASTOBJ reference and should be released using
+ * #ASTOBJ_UNREF(iface, ast_smdi_interface_destroy).
+ */
+extern struct ast_smdi_interface *ast_smdi_interface_find(const char *iface_name)
+{
+	return (ASTOBJ_CONTAINER_FIND(&smdi_ifaces,iface_name));
+}
+
+static struct ast_smdi_switchtype *ast_smdi_switchtype_find(const char *switchtype_name)
+{
+	return (ASTOBJ_CONTAINER_FIND(&smdi_switchtypes,switchtype_name));
+}
+
+/*! \brief Read an SMDI message.
+ *
+ * \param iface the SMDI interface to read from.
+ *
+ * This function loops and reads from and SMDI interface.  It must be stopped
+ * using pthread_cancel().
+ */
+static void *smdi_read(void *iface_p)
+{
+	struct ast_smdi_interface *iface = iface_p;
+	struct ast_smdi_md_message *md_msg;
+	struct ast_smdi_mwi_message *mwi_msg;
+	char buf[256];
+	char c = '\0';
+	char *cp = NULL;
+	char *msg_postfix = NULL;
+	int i;
+		
+	/* read an smdi message */
+	buf[0] = '\0';
+	while((c = fgetc(iface->file))) {
+		if(strlen(buf) + 1 >= sizeof(buf)) {
+			ast_log(LOG_WARNING, "Read 255 characters on SMDI port %s without finding the start of a message.  Check your SMDI switchtype.\n", iface->name);
+			buf[0] = '\0';
+			/* XXX: We may want to copy the last few chars into the front of the buffer */
+		}
+		strncat(buf, &c, 1);
+
+		/* check if this is the start of a message */
+		if(strstr(buf, iface->switchtype->msg_prefix)) {
+			buf[0] = '\0';
+
+			/* Determine if this is a MD or MWI message */
+			if((c = fgetc(iface->file))) {
+				strncat(buf, &c, 1);
+			} else {
+				ast_log(LOG_ERROR, "Error reading from SMDI interface %s, stopping listner thread\n", iface->name);
+				ASTOBJ_UNREF(iface,ast_smdi_interface_destroy);
+				return NULL;
+			}
+			
+			if((c = fgetc(iface->file))) {
+				strncat(buf, &c, 1);
+			} else {
+				ast_log(LOG_ERROR, "Error reading from SMDI interface %s, stopping listner thread\n", iface->name);
+				ASTOBJ_UNREF(iface,ast_smdi_interface_destroy);
+				return NULL;
+			}
+
+			if(!strcmp(buf, "MD")) {
+				/* this is an MD message */
+				buf[0] = '\0';
+	
+				md_msg = malloc(sizeof(struct ast_smdi_md_message));
+				if(!md_msg) {
+					ast_log(LOG_ERROR, "Error allocating memory for ast_smdi_md_message.  Stopping listner thread.\n");
+					ASTOBJ_UNREF(iface,ast_smdi_interface_destroy);
+					return NULL;
+				}
+				memset(md_msg, 0, sizeof(struct ast_smdi_md_message));
+				
+				ASTOBJ_INIT(md_msg);
+
+				/* read the message desk number */
+				for(i = 0; i < SMDI_MESG_DESK_NUM_LEN; i++) {
+					md_msg->mesg_desk_num[i] = fgetc(iface->file);
+				}
+				md_msg->mesg_desk_num[SMDI_MESG_DESK_NUM_LEN] = '\0';
+
+				/* read the message desk terminal number */
+				for(i = 0; i < SMDI_MESG_DESK_TERM_LEN; i++) {
+					md_msg->mesg_desk_term[i] = fgetc(iface->file);
+				}
+				md_msg->mesg_desk_term[SMDI_MESG_DESK_TERM_LEN] = '\0';
+
+				/* read the message type */
+				md_msg->type = fgetc(iface->file);
+			   
+				/* read the forwarding station number (may be blank) */
+				cp = &md_msg->fwd_st[0];
+				for(i = 0; i < SMDI_MAX_STATION_NUM_LEN + 1; i++) {
+					if((c = fgetc(iface->file)) == ' ') {
+						*cp = '\0';
+						break;
+					}
+
+					if( i >= iface->msdstrip) {
+						*cp = c;
+						cp++;
+					}
+				}
+				/* make sure the value is null terminated, even if this truncates it */
+				md_msg->fwd_st[SMDI_MAX_STATION_NUM_LEN] = '\0';
+				cp = NULL;
+				
+				/* read the calling station number (may be blank) */
+				cp = &md_msg->calling_st[0];
+				for(i = 0; i < SMDI_MAX_STATION_NUM_LEN + 1; i++) {
+					if((c = fgetc(iface->file)) == iface->switchtype->md_cid_end) {
+						*cp = '\0';
+						break;
+					}
+
+					if( i >= iface->msdstrip) {
+						*cp = c;
+						cp++;
+					}
+				}
+				/* make sure the value is null terminated, even if this truncates it */
+				md_msg->calling_st[SMDI_MAX_STATION_NUM_LEN] = '\0';
+				cp = NULL;
+	
+				/* add the message to the message queue */
+				ast_smdi_md_message_push(iface, md_msg);
+				ast_log(LOG_DEBUG, "Recieved SMDI MD message on %s\n", iface->name);
+				
+				ASTOBJ_UNREF(md_msg, ast_smdi_md_message_destroy);
+				msg_postfix = iface->switchtype->md_postfix;
+
+			} else if(!strcmp(buf, "MW")) {
+				/* this is an MWI message */
+				buf[0] = '\0';
+				
+				mwi_msg = malloc(sizeof(struct ast_smdi_mwi_message));
+				if(!mwi_msg) {
+					ast_log(LOG_ERROR, "Error allocating memory for ast_smdi_mwi_message.  Stopping listner thread.\n");
+					ASTOBJ_UNREF(iface,ast_smdi_interface_destroy);
+					return NULL;
+				}
+				memset(mwi_msg, 0, sizeof(struct ast_smdi_mwi_message));
+				
+				ASTOBJ_INIT(mwi_msg);
+
+				/* discard the 'I' (from 'MWI') */
+				fgetc(iface->file);
+				
+				/* read the forwarding station number (may be blank) */
+				cp = &mwi_msg->fwd_st[0];
+				for(i = 0; i < SMDI_MAX_STATION_NUM_LEN + 1; i++) {
+					if((c = fgetc(iface->file)) == ' ') {
+						*cp = '\0';
+						break;
+					}
+
+					if( i >= iface->msdstrip) {
+						*cp = c;
+						cp++;
+					}
+				}
+				/* make sure the station number is null terminated, even if this will truncate it */
+				mwi_msg->fwd_st[SMDI_MAX_STATION_NUM_LEN] = '\0';
+				cp = NULL;
+				
+				/* read the mwi failure cause */
+				for(i = 0; i < SMDI_MWI_FAIL_CAUSE_LEN; i++) {
+					mwi_msg->cause[i] = fgetc(iface->file);
+				}
+				mwi_msg->cause[SMDI_MWI_FAIL_CAUSE_LEN] = '\0';
+
+				/* add the message to the message queue */
+				ast_smdi_mwi_message_push(iface, mwi_msg);
+				ast_log(LOG_DEBUG, "Recieved SMDI MWI message on %s\n", iface->name);
+				
+				ASTOBJ_UNREF(mwi_msg, ast_smdi_mwi_message_destroy);
+				msg_postfix = iface->switchtype->mwi_postfix;
+			} else {
+				ast_log(LOG_ERROR, "Unknown SMDI message type recieved on %s (%s).  Check your SMDI switchtype.\n", iface->name, buf);
+				msg_postfix = NULL;
+			}
+			buf[0] = '\0';
+
+		
+			/* discard the rest of the message to the postfix */
+			while((c = fgetc(iface->file)) && msg_postfix) {
+				if(strlen(buf) + 1 >= sizeof(buf)) {
+					ast_log(LOG_WARNING, "Read 255 characters on SMDI port %s without finding the end of a message.  Check your SMDI switchtype.\n", iface->name);
+					buf[0] = '\0';
+					/* XXX: We may want to copy the last few chars into the front of the buffer */
+				}
+				strncat(buf, &c, 1);
+
+				/* check if this is the end of a message */
+				if(strstr(buf, msg_postfix)) {
+					buf[0] = '\0';
+					break;
+				}
+			}
+		}
+	}
+
+	ast_log(LOG_ERROR, "Error reading from SMDI interface %s, stopping listener thread\n", iface->name);
+	ASTOBJ_UNREF(iface,ast_smdi_interface_destroy);
+	return NULL;
+}
+
+/*! \brief ast_smdi_md_message destructor. */
+void ast_smdi_md_message_destroy(struct ast_smdi_md_message *msg)
+{
+	free(msg);
+}
+
+/*! \brief ast_smdi_mwi_message destructor. */
+void ast_smdi_mwi_message_destroy(struct ast_smdi_mwi_message *msg)
+{
+	free(msg);
+}
+
+/*! \brief ast_smdi_interface destructor. */
+void ast_smdi_interface_destroy(struct ast_smdi_interface *iface)
+{
+	if(iface->thread != AST_PTHREADT_NULL && iface->thread != AST_PTHREADT_STOP) {
+		pthread_cancel(iface->thread);
+		pthread_join(iface->thread, NULL);
+	}
+	
+	iface->thread = AST_PTHREADT_STOP;
+	
+	if(iface->file) 
+		fclose(iface->file);
+	
+	if(iface->switchtype)
+		ASTOBJ_UNREF(iface->switchtype,ast_smdi_switchtype_destroy);
+
+	ASTOBJ_CONTAINER_DESTROYALL(&iface->md_q, ast_smdi_md_message_destroy);
+	ASTOBJ_CONTAINER_DESTROYALL(&iface->mwi_q, ast_smdi_mwi_message_destroy);
+	ASTOBJ_CONTAINER_DESTROY(&iface->md_q);
+	ASTOBJ_CONTAINER_DESTROY(&iface->mwi_q);
+	free(iface);
+
+	ast_mutex_lock(&localuser_lock);
+	localusecnt--;
+	ast_mutex_unlock(&localuser_lock);
+	ast_update_use_count();
+}
+
+/*!
+ * \internal
+ * \brief ast_smdi_switchtype destructor.
+ */
+static void ast_smdi_switchtype_destroy(struct ast_smdi_switchtype *switchtype)
+{
+	free(switchtype);
+}
+
+/*!
+ * \internal
+ * \brief Add a reconized switchtype.
+ * \param name the name of the switchtype.
+ * \param msg_prefix the prefix for SMDI messages.
+ * \param md_postfix the postfix for MD messages.
+ * \param mwi_postfix the postfix for MWI messages.
+ * \param md_cid_end the character that marks the end of the CallerID string.
+ *
+ * This function adds a switchtype to alter the SMDI stack's behavior.
+ *
+ * \return zero on success, -1 on failure.
+ */
+static int ast_smdi_switchtype_add(const char *name, const char *msg_prefix, const char *md_postfix, const char *mwi_postfix, char md_cid_end)
+{
+	struct ast_smdi_switchtype *switchtype;
+
+	/* create a switchtype object */
+	switchtype = malloc(sizeof(struct ast_smdi_switchtype));
+	if(!switchtype) 
+		return -1;
+
+	memset(switchtype, 0, sizeof(struct ast_smdi_switchtype));
+	ASTOBJ_INIT(switchtype);
+
+	/* fill in our switchtype object */
+	ast_copy_string(switchtype->name, name, sizeof(switchtype->name));
+	ast_copy_string(switchtype->msg_prefix, msg_prefix, sizeof(switchtype->msg_prefix));
+	ast_copy_string(switchtype->md_postfix, md_postfix, sizeof(switchtype->md_postfix));
+	ast_copy_string(switchtype->mwi_postfix, mwi_postfix, sizeof(switchtype->mwi_postfix));
+	switchtype->md_cid_end = md_cid_end;
+
+	ASTOBJ_CONTAINER_LINK(&smdi_switchtypes,switchtype);
+	ASTOBJ_UNREF(switchtype,ast_smdi_switchtype_destroy);
+
+	return 0;
+}
+
+/*!
+ * \internal
+ * \brief Load and reload SMDI configuration.
+ * \param reload this should be 1 if we are reloading and 0 if not.
+ *
+ * This function loads/reloads the SMDI configuration and starts and stops
+ * interfaces accordingly.
+ *
+ * \return zero on success, -1 on failure, and 1 if no smdi interfaces were started.
+ */
+static int smdi_load(int reload)
+{
+	struct ast_config *conf;
+	struct ast_variable *v;
+	struct ast_smdi_interface *iface = NULL;
+	int res = 0;
+
+	/* Config options */
+	speed_t baud_rate = B9600;     /* 9600 baud rate */
+	tcflag_t paritybit = PARENB;   /* even parity checking */
+	tcflag_t charsize = CS7;       /* seven bit characters */
+	int stopbits = 0;              /* One stop bit */
+	
+	int msdstrip = 0;              /* strip zero digits */
+	
+	struct ast_smdi_switchtype *switchtype = ast_smdi_switchtype_find("default");
+				
+
+	/* Make sure the default switch type exists, if it does not we have serious problems. */
+	if(!switchtype) {
+		ast_log(LOG_ERROR, "default switchtype not found\n");
+		res = -1;
+	}
+
+	conf = ast_config_load(config_file);
+
+	if(!conf) {
+		ast_log(LOG_ERROR, "Unable to load config %s\n", config_file);
+		return -1;
+	}
+
+	/* Mark all interfaces that we are listening on.  We will unmark them
+	 * as we find them in the config file, this way we know any interfaces
+	 * still marked after we have finished parsing the config file should
+	 * be stopped.
+	 */
+	if(reload)
+		ASTOBJ_CONTAINER_MARKALL(&smdi_ifaces);
+
+	v = ast_variable_browse(conf, "interfaces");
+	while(v) {
+		if(!strcasecmp(v->name, "switchtype")) {
+			if(switchtype)
+				ASTOBJ_UNREF(switchtype,ast_smdi_switchtype_destroy);
+			switchtype = ast_smdi_switchtype_find(v->value);
+			if(!switchtype) {
+				ast_log(LOG_NOTICE, "Invalid switchtype '%s' specified in %s (line %d), using default\n", v->value, config_file, v->lineno);
+				switchtype = ast_smdi_switchtype_find("default");
+
+				/* if we don't have the default switchtype we have a serious problem */
+				if(!switchtype) {
+					ast_log(LOG_ERROR, "default switchtype not found\n");
+					res = -1;
+				}
+			}
+		} else if(!strcasecmp(v->name, "baudrate")) {
+			if(!strcasecmp(v->value, "9600")) {
+				baud_rate = B9600;
+			} else if(!strcasecmp(v->value, "4800")) {
+				baud_rate = B4800;
+			} else if(!strcasecmp(v->value, "2400")) {
+				baud_rate = B2400;
+			} else if(!strcasecmp(v->value, "1200")) {
+				baud_rate = B1200;
+			} else {
+				ast_log(LOG_NOTICE, "Invalid baud rate '%s' specified in %s (line %d), using default\n", v->value, config_file, v->lineno);
+				baud_rate = B9600;
+			}
+		} else if(!strcasecmp(v->name, "msdstrip")) {
+			if(!sscanf(v->value, "%d", &msdstrip)) {
+				ast_log(LOG_NOTICE, "Invalid msdstrip value in %s (line %d), using default\n", config_file, v->lineno);
+				msdstrip = 0;
+			} else if(0 > msdstrip || msdstrip > 9) {
+				ast_log(LOG_NOTICE, "Invalid msdstrip value in %s (line %d), using default\n", config_file, v->lineno);
+				msdstrip = 0;
+			}
+		} else if(!strcasecmp(v->name, "paritybit")) {
+			if(!strcasecmp(v->value, "even")) {
+				paritybit = PARENB;
+			} else if(!strcasecmp(v->value, "odd")) {
+				paritybit = PARENB | PARODD;
+			} else if(!strcasecmp(v->value, "none")) {
+				paritybit = ~PARENB;
+			} else {
+				ast_log(LOG_NOTICE, "Invalid parity bit setting in %s (line %d), using default\n", config_file, v->lineno);
+				paritybit = PARENB;
+			}
+		} else if(!strcasecmp(v->name, "charsize")) {
+			if(!strcasecmp(v->value, "7")) {
+				charsize = CS7;
+			} else if(!strcasecmp(v->value, "8")) {
+				charsize = CS8;
+			} else {
+				ast_log(LOG_NOTICE, "Invalid character size setting in %s (line %d), using default\n", config_file, v->lineno);
+				charsize = CS7;
+			}
+		} else if(!strcasecmp(v->name, "twostopbits")) {
+			stopbits = ast_true(v->name);
+		} else if(!strcasecmp(v->name, "smdiport")) {
+			if(reload) {
+				/* we are reloading, check if we are already
+				 * monitoring this interface, if we are we do
+				 * not want to start it again.  This also has
+				 * the side effect of not updating different
+				 * setting for the serial port, but it should
+				 * be trivial to rewrite this section so that
+				 * options on the port are changed without
+				 * restarting the interface.  Or the interface
+				 * could be restarted with out emptying the
+				 * queue. */
+				iface = ASTOBJ_CONTAINER_FIND(&smdi_ifaces, v->value);
+				if(iface) {
+					ast_log(LOG_NOTICE, "SMDI interface %s already running, not restarting\n", iface->name);
+					ASTOBJ_UNMARK(iface);
+					ASTOBJ_UNREF(iface, ast_smdi_interface_destroy);
+					v = v->next;
+					continue;
+				}
+			}
+							
+			iface = malloc(sizeof(struct ast_smdi_interface));
+
+			ASTOBJ_INIT(iface);
+			ASTOBJ_CONTAINER_INIT(&iface->md_q);
+			ASTOBJ_CONTAINER_INIT(&iface->mwi_q);
+			iface->md_q.head = NULL;
+			iface->mwi_q.head = NULL;
+			iface->thread = AST_PTHREADT_NULL;
+			memset(&iface->mode, 0, sizeof(iface->mode));
+
+			ast_copy_string(iface->name, v->value, sizeof(iface->name));
+
+			iface->switchtype = switchtype;
+
+			iface->file = fopen(iface->name, "r+");
+			if(!(iface->file)) {
+				ast_log(LOG_ERROR, "Error opening SMDI interface %s (%s)\n", iface->name, strerror(errno));
+				ASTOBJ_UNREF(iface, ast_smdi_interface_destroy);
+		                v = v->next;
+				continue;
+			}
+			iface->fd = fileno(iface->file);
+
+			/* Set the proper attributes for our serial port. */
+
+			/* get the current attributes from the port */
+			if(tcgetattr(iface->fd, &iface->mode)) {
+				ast_log(LOG_ERROR, "Error getting atributes of %s (%s)\n", iface->name, strerror(errno));
+				ASTOBJ_UNREF(iface, ast_smdi_interface_destroy);
+				v = v->next;
+				continue;
+			}
+
+			/* set the desired speed */
+			if(cfsetispeed(&iface->mode, baud_rate) || cfsetospeed(&iface->mode, baud_rate)) {
+				ast_log(LOG_ERROR, "Error setting baud rate on %s (%s)\n", iface->name, strerror(errno));
+				ASTOBJ_UNREF(iface, ast_smdi_interface_destroy);
+				v = v->next;
+				continue;
+			}
+			
+			/* set the stop bits */
+			if(stopbits)
+			   iface->mode.c_cflag = iface->mode.c_cflag | CSTOPB;   /* set two stop bits */
+			else
+			   iface->mode.c_cflag = iface->mode.c_cflag & ~CSTOPB;  /* set one stop bit */
+
+			/* set the parity */
+			iface->mode.c_cflag = (iface->mode.c_cflag & ~PARENB & ~PARODD) | paritybit;
+
+			/* set the character size */
+			iface->mode.c_cflag = (iface->mode.c_cflag & ~CSIZE) | charsize;
+			
+			/* commit the desired attributes */
+			if(tcsetattr(iface->fd, TCSAFLUSH, &iface->mode)) {
+				ast_log(LOG_ERROR, "Error setting attributes on %s (%s)\n", iface->name, strerror(errno));
+				ASTOBJ_UNREF(iface, ast_smdi_interface_destroy);
+				v = v->next;
+				continue;
+			}
+
+			/* set the msdstrip */
+			iface->msdstrip = msdstrip;
+
+                        /* start the listner thread */
+			if(option_verbose > 2)
+				ast_verbose(VERBOSE_PREFIX_3 "Starting SMDI monitor thread for %s\n", iface->name);
+			if(ast_pthread_create(&iface->thread, NULL, smdi_read, iface)) {
+				ast_log(LOG_ERROR, "Error starting SMDI monitor thread for %s\n", iface->name);
+				ASTOBJ_UNREF(iface, ast_smdi_interface_destroy);
+		                v = v->next;
+				continue;
+			}
+			ASTOBJ_CONTAINER_LINK(&smdi_ifaces, iface);
+			ast_mutex_lock(&localuser_lock);
+			localusecnt++;
+			ast_mutex_unlock(&localuser_lock);
+			ast_update_use_count();
+		} else {
+			ast_log(LOG_NOTICE, "Ignoring unknown option %s in %s\n", v->name, config_file);
+		}
+		v = v->next;
+	}
+	ast_config_destroy(conf);
+
+	/* Prune any interfaces we should no longer monitor. */
+	if(reload)
+		ASTOBJ_CONTAINER_PRUNE_MARKED(&smdi_ifaces, ast_smdi_interface_destroy);
+	
+	ASTOBJ_CONTAINER_RDLOCK(&smdi_ifaces);
+	if(!smdi_ifaces.head)
+		res = 1;
+	ASTOBJ_CONTAINER_UNLOCK(&smdi_ifaces);
+			
+	if(switchtype)
+		ASTOBJ_UNREF(switchtype,ast_smdi_switchtype_destroy);
+
+	return res;
+}
+
+
+char *description(void)
+{
+	return (char *)tdesc;
+}
+
+int load_module(void)
+{
+	int res;
+
+	/* initilize our containers */
+	ASTOBJ_CONTAINER_INIT(&smdi_ifaces);
+	smdi_ifaces.head = NULL;
+
+	ASTOBJ_CONTAINER_INIT(&smdi_switchtypes);
+	smdi_switchtypes.head = NULL;
+
+	/* build our list of switchtypes */
+	/* XXX maybe check return values here */
+	ast_smdi_switchtype_add("default", "\r\n", "\r\n\x19", "\r\n\x7F\x7F\x19", ' ');
+	ast_smdi_switchtype_add("dms10", "\n\n", "\n\x19", "\n\x19", '\n');
+	ast_smdi_switchtype_add("g3", "\n\n", "\n\n", "\n\n", '\n');
+
+	/* load the config and start the listener threads*/
+	res = smdi_load(0);
+	if(res < 0) {
+		return res;
+	} else if(res == 1) {
+		ast_log(LOG_WARNING, "No SMDI interfaces are available to listen on, not starting SDMI listener.\n");
+		return 0;
+	}
+
+	return 0;
+}
+
+int unload_module(void)
+{
+	/* this destructor stops any running smdi_read threads */
+	ASTOBJ_CONTAINER_DESTROYALL(&smdi_ifaces,ast_smdi_interface_destroy);
+	ASTOBJ_CONTAINER_DESTROY(&smdi_ifaces);
+
+	ASTOBJ_CONTAINER_DESTROYALL(&smdi_switchtypes,ast_smdi_switchtype_destroy);
+	ASTOBJ_CONTAINER_DESTROY(&smdi_switchtypes);
+	return 0;
+}
+
+int reload(void)
+{
+	int res;
+	res = smdi_load(1);
+
+	if(res < 0) {
+		return res;
+	} else if(res == 1) {
+		ast_log(LOG_WARNING, "No SMDI interfaces were specified to listen on, not starting SDMI listener.\n");
+		return 0;
+	}
+
+	/* notify the listner thread? */
+	return 0;
+}
+
+int usecount(void)
+{
+	int res;
+	STANDARD_USECOUNT(res);
+	return res;
+}
+
+char *key()
+{
+	return ASTERISK_GPL_KEY;
+}
+
